/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  EdgeGame								//
//  --------								//
//									//
//  Specific game engine implementation for EDGE.			//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <klocale.h>
#include <kglobal.h>

#include <qwidget.h>
#include <qdir.h>
#include <qfile.h>
#include <qurl.h>
#include <qregexp.h>
#include <qfileinfo.h>

#include "gamemanager.h"
#include "genericgame.h"
#include "edgedisplayoptions.h"
#include "edgesoundoptions.h"
#include "edgeenemiesoptions.h"

#include "edgegame.h"
#include "edgegame.moc"

//////////////////////////////////////////////////////////////////////////
//									//
//  Engine information							//
//									//
//////////////////////////////////////////////////////////////////////////

class EdgeGameInfo
{
public:
	static const GameManager::gameInfo myinfo;

protected:
	static GenericGame *create(QWidget *parent);

private:
	static EdgeGameInfo initobj;
	EdgeGameInfo();
};


const GameManager::gameInfo EdgeGameInfo::myinfo =
{
	"edge",						// key
	"EDGE",						// name
	"http://edge.sourceforge.net",			// url
	i18n("This engine cannot play back demo files in the old LMP format."),
							// notes
	"edge:gledge:gledge32:gledge64",		// execname
	QString::null,					// execpaths
	QString::null,					// datapaths
	&EdgeGameInfo::create				// createfunc
};


EdgeGameInfo EdgeGameInfo::initobj;


EdgeGameInfo::EdgeGameInfo()
{
	GameManager::self()->add(&EdgeGameInfo::myinfo);
}


GenericGame *EdgeGameInfo::create(QWidget *parent)
{
	kdDebug(0) << k_funcinfo << endl;
	GenericGame *g = new EdgeGame(parent);
	return (g);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

EdgeGame::EdgeGame(QWidget *parent)
	: GenericGame(parent)
{
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Option dialogues							//
//									//
//////////////////////////////////////////////////////////////////////////

void EdgeGame::slotDisplayOptions()
{
	EdgeDisplayOptions e(this,displaysize,displayoption);

	e.show();
	if (e.exec())
	{
		displayoption = e.options();
		displaysize = e.size();
	}
}


void EdgeGame::slotSoundOptions()
{
	EdgeSoundOptions e(this,soundoption);

	e.show();
	if (e.exec()) soundoption = e.options();
}


void EdgeGame::slotEnemiesOptions()
{
	EdgeEnemiesOptions e(this,enemiesoption);

	e.show();
	if (e.exec()) enemiesoption = e.options();
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Engine command line							//
//									//
//////////////////////////////////////////////////////////////////////////

QString EdgeGame::optionsDisplay(QStringList &cmd)
{
	GenericGame::optionsDisplay(cmd);
	if (displayoption & DisplayOptions::Windowed) cmd += "-windowed";
	else cmd += "-fullscreen";

	return (QString::null);
}


QString EdgeGame::playStartLevel(QStringList &cmd,const QString &warp,int skill)
{
	cmd += "-warp";
	if (warp.startsWith("E"))			// Doom 1 level format
	{
		int e,m;
		sscanf(warp.upper().ascii(),"E%dM%d:",&e,&m);
		cmd += QString::number(e);
		cmd += QString::number(m);
	}
	else cmd += warp;				// plain Doom 2 format
							// skill in EDGE is 1..5
	cmd += "-skill"; cmd += QString::number(skill+1);
	return (QString::null);
}


QString EdgeGame::optionsSound(QStringList &cmd)
{
	if (soundoption==SoundOptions::NoSound) cmd += "-nosound";
	else
	{
		if (soundoption & SoundOptions::NoMusic) cmd += "-nomusic";
		if (soundoption & SoundOptions::CDMusic) cmd += "-cdmusic";
	}
	return (QString::null);
}


QString EdgeGame::optionsEnemies(QStringList &cmd)
{
	if (enemiesoption & EnemiesOptions::None) cmd += "-nomonsters";
	else
	{
		if (enemiesoption & EnemiesOptions::Respawn) cmd += "-respawn";
		if (enemiesoption & EnemiesOptions::Fast) cmd += "-fast";
	}

	if (enemiesoption & EnemiesOptions::Infight) cmd += "-infight";
	if (enemiesoption & EnemiesOptions::MoreBlood) cmd += "-blood";
	return (QString::null);
}


QString EdgeGame::getPlayCommand(QStringList &cmd)
{
	kdDebug(0) << k_funcinfo << endl;

	QUrl u(cmd.first());				// get executable path
	cmd += "-game";
	cmd += u.dirPath();				// use it to locate EDGE.WAD

	return (GenericGame::getPlayCommand(cmd));
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Saved games and information						//
//									//
//////////////////////////////////////////////////////////////////////////

QStringList EdgeGame::findSavedGames()
{
	kdDebug(0) << k_funcinfo << endl;
	QStringList l;

	QDir d = QDir::home();
	if (!d.cd(".edge/savegame"))
	{
		kdDebug(0) << k_funcinfo << "no saved game directory" << endl;
		return (l);
	}

	d.setFilter(QDir::Files+QDir::Readable);
	d.setNameFilter("save[0-9][0-9][0-9][0-9].esg");
	d.setSorting(QDir::Name);			// name filter is a shell glob

	QRegExp rx("save(....)\\.esg");			// but this is a regexp
	QStringList entries = d.entryList();
	for (QStringList::const_iterator it = entries.begin(); it!=entries.end(); ++it)
	{
		const QString entry = (*it);
		if (!rx.exactMatch(entry)) continue;	// should always match
		const QString idx = rx.cap(1);		// number captured from there

		QString buf;				// file "0001" is index 0
		buf.sprintf("%04d",idx.toInt(NULL,10)-1);

		QFileInfo fi(d,entry);			// show game saved time
		l.append(QString("%1: %2").arg(buf,KGlobal::locale()->formatDateTime(fi.lastModified(),true,true)));
	}

	kdDebug(0) << k_funcinfo << "returning " << l.join(",") << endl;
	return (l);
}


const GameManager::gameInfo *EdgeGame::gameInfo()
{
	return (&EdgeGameInfo::myinfo);
}


QString EdgeGame::demoFormat() const			// EDGE can't play LMP demos
{
	return ("EDM");					// but can play its own
}
