/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  GameManager								//
//  -----------								//
//									//
//  This class maintains and searches the list of supported game	//
//  engines, and information about each of them.  There is only one	//
//  unique instance.							//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <klocale.h>
#include <kstandarddirs.h>

#include <qstring.h>
#include <qstringlist.h>
#include <qptrlist.h>
#include <qdir.h>

#include "gamemanager.h"

//////////////////////////////////////////////////////////////////////////
//									//
//  Common text strings							//
//									//
//////////////////////////////////////////////////////////////////////////

							// used for WAD and PWAD
const QString GameManager::wadFilterText = i18n("*.WAD *.wad|Doom WAD Files (*.wad)\n"
						"*.ZIP *.zip|Zip Archives (*.zip)\n"
						"*|All Files");
const QString GameManager::wadCaptionText = i18n("Select WAD");

const QString GameManager::lmpFilterText = i18n("*.LMP *.lmp|Doom LMP Files (*.lmp)\n"
						"*.ZIP *.zip|Zip Archives (*.zip)\n"
						"*.EDM *.edm|EDGE Demo Files (*.edm)\n"
						"*|All Files");
const QString GameManager::lmpCaptionText = i18n("Select Demo");

//////////////////////////////////////////////////////////////////////////
//									//
//  Standard search paths						//
//									//
//////////////////////////////////////////////////////////////////////////

static const QString execPaths = "/usr/games/bin:/usr/local/bin:/usr/bin";
static const QString dataPaths = "/usr/share/games/doom-data:/usr/share/games/doom";

//////////////////////////////////////////////////////////////////////////
//									//
//  Instance and access							//
//									//
//////////////////////////////////////////////////////////////////////////

static GameManager *instance = NULL;

GameManager *GameManager::self()
{
	if (instance==NULL) instance = new GameManager();
	return (instance);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

GameManager::GameManager()
{
	games = new QPtrList<GameManager::gameInfo>;
}


GameManager::~GameManager()
{
	delete games;
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Register a new engine, done for each at initialisation time		//
//									//
//////////////////////////////////////////////////////////////////////////

void GameManager::add(const GameManager::gameInfo *inf)
{
	kdDebug(0) << k_funcinfo << "add '" << inf->key << "' = '" << inf->name << "'" << endl;

	games->append(inf);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Create an engine implementation object				//
//									//
//////////////////////////////////////////////////////////////////////////

GenericGame *GameManager::create(const QString &key,QWidget *parent)
{
	kdDebug(0) << k_funcinfo << "for '" << key << "'" << endl;

	const GameManager::gameInfo *inf = GameManager::findInfo(key);
	if (inf==NULL) return (NULL);
	return ((inf->create)(parent));
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Searching and iterating over the list				//
//									//
//////////////////////////////////////////////////////////////////////////

const GameManager::gameInfo *GameManager::firstInfo() const
{
	return (games->first());
}


const GameManager::gameInfo *GameManager::nextInfo() const
{
	return (games->next());
}


const GameManager::gameInfo *GameManager::findInfo(const QString &key) const
{
	for (const GameManager::gameInfo *it = games->first(); it!=NULL; it = games->next())
	{
		if (key==it->key) return (it);
	}
	return (NULL);
}


const GameManager::gameInfo *GameManager::findNamed(const QString &name) const
{
	for (const GameManager::gameInfo *it = games->first(); it!=NULL; it = games->next())
	{
		if (name==it->name) return (it);
	}
	return (NULL);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Searching for files							//
//									//
//////////////////////////////////////////////////////////////////////////

const QString GameManager::findExecFor(const GameManager::gameInfo *g) const
{
	QString exe = QString::null;

	const QStringList names = QStringList::split(':',g->execname);
	for (QStringList::const_iterator it = names.begin(); it!=names.end(); ++it)
	{						// try all possible names
		if (!g->execpaths.isNull()) exe = KStandardDirs::findExe(*it,g->execpaths);
		if (exe.isNull()) exe = KStandardDirs::findExe(*it,execPaths);
		if (exe.isNull()) exe = KStandardDirs::findExe(*it);
		if (!exe.isNull()) break;
	}

	kdDebug(0) << k_funcinfo << "for " << g->name << " = " << exe << endl;
	return (exe);
}


const QString GameManager::findDataFor(const GameManager::gameInfo *g) const
{
	QString dir = QString::null;
	QStringList dl;

	if (!g->datapaths.isNull()) dl += QStringList::split(':',g->datapaths,true);
	dl += QStringList::split(':',dataPaths,true);

	for ( QStringList::ConstIterator it = dl.begin(); it!=dl.end(); ++it)
	{						// try all possible locations
		QDir d(*it);
		if (!d.exists()) continue;

		d.setFilter(QDir::Files+QDir::Readable);
		d.setNameFilter("*.wad;*.WAD");
		if (d.count()>0)			// look for any WAD files
		{
			dir = d.absPath();
			break;
		}
	}

	kdDebug(0) << k_funcinfo << "for " << g->name << " = " << dir << endl;
	return (dir);
}
