/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  GenericGame								//
//  -----------								//
//									//
//  This is the base class of the implementation classes for all the	//
//  game engines that are supported.  It provides the basic framework	//
//  used by all of them, which can be augumented or replaced for each	//
//  specific engine as necessary.					//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <kdialog.h>
#include <klocale.h>
#include <kfile.h>
#include <kfiledialog.h>
#include <kaction.h>
#include <kxmlguiclient.h>
#include <kxmlguifactory.h>
#include <kiconloader.h>
#include <kinputdialog.h>
#include <kprotocolinfo.h>

#include <qlayout.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qgroupbox.h>
#include <qlistbox.h>
#include <qstringlist.h>
#include <qcheckbox.h>
#include <qpopupmenu.h>
#include <qsize.h>
#include <qurl.h>
#include <qvalidator.h>

#include <qdir.h>
#include <qfile.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "wad.h"
#include "pwadlist.h"
#include "remotefiles.h"
#include "infodisplay.h"
#include "demorequester.h"
#include "listparser.h"
#include "gamemanager.h"

#include "genericgame.h"
#include "genericgame.moc"

//  We'd like to use these functions, but the header file that defines them is
//  only used internally and not installed by KDE!

#ifndef HAVE_KRECENTDIRS_H
class KRecentDirs
{
public:	static QString dir(const QString &fileClass);
	static void add(const QString &fileClass,const QString &directory);
};
#endif

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

GenericGame::GenericGame(QWidget *parent)
	: QWidget(parent)
{
	kdDebug(0) << k_funcinfo << endl;

	const int mh = KDialog::marginHint();		// conveniences, use them a lot
	const int sh = KDialog::spacingHint();

	QVBoxLayout *l1 = new QVBoxLayout(this,0,sh);	// main layout
							// "WAD Files" group
	QGroupBox *g = new QGroupBox(i18n("WAD Files"),this);
	QGridLayout *l2 = new QGridLayout(g,7,5,mh,sh);	// its internal layout

	mainwad = new KURLRequester(g);			// "Main" file requester
	mainwad->setMode(KFile::File+KFile::ExistingOnly+KFile::LocalOnly);
	mainwad->setFilter(GameManager::wadFilterText);
	mainwad->fileDialog()->setCaption(GameManager::wadCaptionText);
	l2->addMultiCellWidget(mainwad,1,1,2,3);

	QLabel *lab = new QLabel(i18n("Base:"),g);	// its text label
	l2->addWidget(lab,1,1,Qt::AlignRight);
	lab->setBuddy(mainwad);

	pwads = new PWadList(g);			// "Additional" list and buttons
	l2->addMultiCellWidget(pwads,2,5,2,3);
	l2->setRowStretch(2,1);

	lab = new QLabel(i18n("Additional:"),g);	// its text label
	l2->addWidget(lab,2,1,Qt::AlignRight+Qt::AlignTop);
	lab->setBuddy(pwads);

	l2->setRowSpacing(0,mh); l2->setRowSpacing(6,sh);
	l2->setColSpacing(0,sh); l2->setColSpacing(4,sh); l2->setColStretch(2,9);

	g->adjustSize();
	l1->addWidget(g,2);

	g = new QGroupBox(i18n("Play"),this);		// "Play" group
	l2 = new QGridLayout(g,8,6,mh,sh);		// its internal layout

	loadgame = new QComboBox(g);
	l2->addMultiCellWidget(loadgame,1,1,2,4);

	lab = new QLabel(i18n("Load Saved Game:"),g);	// its text label
	l2->addWidget(lab,1,1,Qt::AlignRight);
	lab->setBuddy(loadgame);

	l2->setRowSpacing(2,sh);			// spacing row

	demolmp = new DemoRequester(g);			// "Demo" file requester
	demolmp->setMode(KFile::File+KFile::ExistingOnly);
	demolmp->setFilter(GameManager::lmpFilterText);
	demolmp->fileDialog()->setCaption(GameManager::lmpCaptionText);

        // Check whether saved last location supports listing.  If it doesn't,
	// fileDialog()->setURL() will put up a "Malformed URL" message box.
	KURL last = KRecentDirs::dir(":demos");
        if (last.isValid() && KProtocolInfo::supportsListing(last)) demolmp->fileDialog()->setURL(last);
	l2->addMultiCellWidget(demolmp,3,3,2,2);

	QPushButton *p = new QPushButton(KGlobal::iconLoader()->loadIcon("clear_left",KIcon::Small),QString::null,g);
	connect(p,SIGNAL(clicked()),SLOT(slotClearDemo()));
	l2->addWidget(p,3,3,Qt::AlignLeft);		// its clear button

	lab = new QLabel(i18n("Run Demo:"),g);		// its text label
	l2->addWidget(lab,3,1,Qt::AlignRight);
	lab->setBuddy(demolmp);

	demoinfo = new QPushButton(KGlobal::iconLoader()->loadIcon("info",KIcon::Small),i18n("Demo Info..."),g);
	connect(demoinfo,SIGNAL(clicked()),SLOT(slotDemoInfo()));
	l2->addWidget(demoinfo,3,4);			// its information button

	l2->setRowSpacing(4,sh);			// spacing row

	startlevel = new QListBox(g);			// "Episode/Level" list box
	startlevel->setSelectionMode(QListBox::Single);
	startlevel->setHScrollBarMode(QScrollView::AlwaysOff);
	startlevel->setVScrollBarMode(QScrollView::AlwaysOn);
	startlevel->setMinimumHeight(80);
	l2->addMultiCellWidget(startlevel,5,5,2,4);
	l2->setRowStretch(5,1);
							// its text label
	lab = new QLabel(i18n("Start Episode/Level:"),g);
	l2->addWidget(lab,5,1,Qt::AlignRight+Qt::AlignTop);
	lab->setBuddy(startlevel);

	skill = new QComboBox(g);			// "Difficulty" combo box
	skill->insertItem("I'm Too Young To Die");
	skill->insertItem("Hey, Not Too Rough");
	skill->insertItem("Hurt Me Plenty");
	skill->insertItem("Ultra-Violence");
	skill->insertItem("Nightmare!");
	skill->setCurrentItem(2);
	l2->addMultiCellWidget(skill,6,6,2,4);

	lab = new QLabel(i18n("Difficulty:"),g);	// its text label
	l2->addWidget(lab,6,1,Qt::AlignRight);
	lab->setBuddy(skill);

	l2->setRowSpacing(0,mh); l2->setRowSpacing(7,sh); l2->setRowStretch(5,9);
	l2->setColSpacing(0,sh); l2->setColSpacing(4,sh); l2->setColStretch(2,9);

	g->adjustSize();
	l1->addWidget(g,2);

	connect(mainwad,SIGNAL(textChanged(const QString &)),SLOT(slotSelectedWad(const QString &)));
	connect(loadgame,SIGNAL(activated(int)),SLOT(slotOptionsChanged()));
	connect(demolmp,SIGNAL(textChanged(const QString &)),SLOT(slotSelectedLmp(const QString &)));
	connect(startlevel,SIGNAL(selectionChanged()),SLOT(slotOptionsChanged()));
	connect(skill,SIGNAL(activated(int)),SLOT(slotOptionsChanged()));

	advdisplay = advsound = advenemies = NULL;

	displayoption = DisplayOptions::Null;
	displaysize = QSize(800,600);
	soundoption = SoundOptions::Null;
	enemiesoption = EnemiesOptions::Null;
	wadvalid = false;
	loadsaved = -1;

	slotSelectedWad();
}


GenericGame::~GenericGame()
{
	kdDebug(0) << k_funcinfo << endl;
#ifdef USE_ACTIONLIST
	KXMLGUIClient *client = static_cast<KXMLGUIClient *>(this->parent());
	if (client!=NULL) client->unplugActionList("options_advanced");
#else
	if (advdisplay!=NULL) advdisplay->unplugAll();
	if (advsound!=NULL) advdisplay->unplugAll();
	if (advenemies!=NULL) advenemies->unplugAll();
#endif
	delete advdisplay;
	delete advsound;
	delete advenemies;
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Actions								//
//									//
//////////////////////////////////////////////////////////////////////////

void GenericGame::setupActions(KXMLGUIClient *client)
{
	advdisplay = new KAction(i18n("Display..."),"kadoom_screen",0,this,SLOT(slotDisplayOptions()),
				 client->actionCollection(),"options_display");

	advsound = new KAction(i18n("Sound..."),"multimedia",0,this,SLOT(slotSoundOptions()),
				 client->actionCollection(),"options_sound");

	advenemies = new KAction(i18n("Enemies..."),"kadoom_enemies",0,this,SLOT(slotEnemiesOptions()),
				 client->actionCollection(),"generic_enemies");

	advextra = new KAction(i18n("Advanced..."),"run",0,this,SLOT(slotExtraOptions()),
				 client->actionCollection(),"options_extra");

#ifdef USE_ACTIONLIST					// didn't seem to work - why not?
	QPtrList<KAction> actions;
	actions.append(advdisplay);
	actions.append(advsound);
	actions.append(advenemies);
	actions.append(advextra);

	client->unplugActionList("options_advanced");
	client->plugActionList("options_advanced",actions);
#else
	QPopupMenu *optmenu = static_cast<QPopupMenu *>(client->factory()->container("options",client));
	if (optmenu==NULL) optmenu = static_cast<QPopupMenu *>(client->factory()->container("settings",client));

	advdisplay->plug(optmenu,0);
	advsound->plug(optmenu,1);
	advenemies->plug(optmenu,2);
	advextra->plug(optmenu,3);
	optmenu->insertSeparator(4);
#endif
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Reading and saving options						//
//									//
//////////////////////////////////////////////////////////////////////////

void GenericGame::readOptions(KConfig *config)
{
	const QString name = configName();
	kdDebug(0) << k_funcinfo << "for '" << name << "'" << endl;

	config->setGroup(name+" Display");

	QSize size = config->readSizeEntry("Size");
	if (size.isValid() && !size.isNull()) displaysize = size;
	displayoption = static_cast<DisplayOptions::Option>(config->readNumEntry("Options"));

	config->setGroup(name+" Sound");
	soundoption = static_cast<SoundOptions::Option>(config->readNumEntry("Options"));

	config->setGroup(name+" Enemies");
	enemiesoption = static_cast<EnemiesOptions::Option>(config->readNumEntry("Options"));

	config->setGroup(name+" Wads");
	mainwad->setURL(config->readPathEntry("Main"));
	QStringList l = config->readPathListEntry("Add");
	for (QStringList::const_iterator it = l.begin(); it!=l.end(); ++it)
	{
		pwads->listBox()->insertItem(*it);
	}

	config->setGroup(name+" Game");
	demolmp->setURL(config->readPathEntry("Demo"));
	loadsaved = config->readNumEntry("Load",-1);

	int s = config->readNumEntry("Start",-1);
	if (s>0) startlevel->setSelected(s,true);
	s = config->readNumEntry("Skill",-1);
	if (s>=0) skill->setCurrentItem(s);

	config->setGroup(name+" Advanced");
	extraargs = config->readEntry("Extra");
}


void GenericGame::saveOptions(KConfig *config)
{
	const QString name = configName();
	kdDebug(0) << k_funcinfo << "for '" << name << "'" << endl;

	config->setGroup(name+" Display");
	config->writeEntry("Size",displaysize);
	config->writeEntry("Options",static_cast<int>(displayoption));

	config->setGroup(name+" Sound");
	config->writeEntry("Options",static_cast<int>(soundoption));

	config->setGroup(name+" Enemies");
	config->writeEntry("Options",static_cast<int>(enemiesoption));

	config->setGroup(name+" Wads");
	config->writePathEntry("Main",mainwad->url());
	QStringList l;
	for (unsigned int i = 0; i<pwads->listBox()->count(); ++i)
	{
		l += pwads->listBox()->item(i)->text();
	}
	config->writePathEntry("Add",l);

	config->setGroup(name+" Game");
	config->writePathEntry("Demo",demolmp->url());
	config->writeEntry("Load",loadgame->currentItem());
	config->writeEntry("Start",startlevel->index(startlevel->selectedItem()));
	config->writeEntry("Skill",skill->currentItem());

	config->setGroup(name+" Advanced");
	config->writeEntry("Extra",extraargs);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Action slots and option dialogues					//
//									//
//////////////////////////////////////////////////////////////////////////

void GenericGame::slotClearDemo()
{
	demolmp->clear();
	slotOptionsChanged();
}


void GenericGame::slotDemoInfo()
{
	InfoDisplay *inf = new InfoDisplay(demolmp->url());
	inf->showInfo();
}


void GenericGame::slotDisplayOptions()
{
	DisplayOptions e(this,displaysize,displayoption);

	e.show();
	if (e.exec())
	{
		displayoption = e.options();
		displaysize = e.size();
	}
}


void GenericGame::slotSoundOptions()
{
	SoundOptions e(this,soundoption);

	e.show();
	if (e.exec()) soundoption = e.options();
}


void GenericGame::slotEnemiesOptions()
{
	EnemiesOptions e(this,enemiesoption);

	e.show();
	if (e.exec()) enemiesoption = e.options();
}


void GenericGame::slotExtraOptions()
{
	QRegExpValidator val(this);			// accept anything, including null
	bool accepted;
	QString e = KInputDialog::getText(i18n("Advanced Options"),
					  i18n("Additional engine command line options:"),
					  extraargs,&accepted,this,NULL,&val);
	if (!accepted) return;				// give up if "Cancel" used
	extraargs = e;					// save the new result
}


void GenericGame::slotOptionsChanged()
{
	const bool loading = (loadgame->currentItem()!=0);
	const bool demoing = !demolmp->url().isEmpty();

	loadgame->setEnabled(!demoing);
	startlevel->setEnabled(wadvalid && !loading && !demoing);
	skill->setEnabled(wadvalid && !loading && !demoing && !startlevel->item(0)->isSelected());
	demolmp->setEnabled(!demoFormat().isNull() && !loading);
	demoinfo->setEnabled(!demoFormat().isNull() && demoing);
}


void GenericGame::slotSelectedWad(const QString &file)
{
	kdDebug(0) << k_funcinfo << "file='" << file << "' type=" << Wad::identify(file) << endl;

	startlevel->clear();
	startlevel->insertItem(i18n("(Introduction)"));
	startlevel->setCurrentItem(0);

	if (!file.isEmpty())
	{
		Wad::Type type = Wad::identify(file);
		wadvalid = (type!=Wad::Unknown && type!=Wad::Invalid);
		if (wadvalid) startlevel->insertStringList(Wad::episodeList(type));
	}

	slotOptionsChanged();
	checkPlayAllowed();
}


void GenericGame::slotSelectedLmp(const QString &file)
{
	kdDebug(0) << k_funcinfo << "file='" << file << endl;

	if (!file.isEmpty())
	{
		QUrl u(file);
		u.setFileName(QString::null);
		KRecentDirs::add(":demos",u);
	}

	slotOptionsChanged();
	checkPlayAllowed();
}


void GenericGame::checkPlayAllowed()
{
	const bool ok = !mainwad->url().isEmpty();
	emit readyToPlay(ok,!demolmp->url().isEmpty());
}


void GenericGame::setWadDataDirectory(const QString &location)
{
	kdDebug(0) << k_funcinfo << "to " << location << endl;
	mainwad->fileDialog()->setURL(location);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Engine command line							//
//									//
//  Each of the functions here can be overridden in a subclass if	//
//  necessary.  They append their arguments and options to the		//
//  command list 'cmd', returning QString::null if everything is OK	//
//  or otherwise an error message text.  getPlayCommand() can be	//
//  reimplemented to take over generation of the entire command line, 	//
//  but unless the engine has a completely nonstandard syntax that	//
//  will not normally be needed.					//
//									//
//////////////////////////////////////////////////////////////////////////

QString GenericGame::optionsDisplay(QStringList &cmd)
{
	cmd += "-width"; cmd += QString::number(displaysize.width());
	cmd += "-height"; cmd += QString::number(displaysize.height());
	if (displayoption & DisplayOptions::Expand2) cmd += "-2";
	else if (displayoption & DisplayOptions::Expand3) cmd += "-3";
	if (displayoption & DisplayOptions::NoAccel) cmd += "-noaccel";
	if (displayoption & DisplayOptions::NoJoy) cmd += "-nojoy";
	if (displayoption & DisplayOptions::NoMouse) cmd += "-nomouse";

	return (QString::null);
}


QString GenericGame::getMainWad(QStringList &cmd,RemoteFiles *files,const KURL &u)
{
	QString wadfile = files->get(u,"WAD");
	if (wadfile.isNull()) return (files->lastError());

	QFileInfo inf = QFileInfo(wadfile);
	if (!inf.isReadable()) return (i18n("The WAD file '<b>%1</b>' was not found or is not readable").arg(wadfile));
	if (Wad::identify(wadfile)==Wad::Invalid) return (i18n("The WAD file '<b>%1</b>' cannot be used").arg(wadfile));

	cmd += "-iwad"; cmd += wadfile;
	return (QString::null);
}


QString GenericGame::getPWads(QStringList &cmd,RemoteFiles *files,QListBox *list)
{
	bool first = true;
	for (unsigned int i = 0; i<list->count(); ++i)
	{
		QString wadfile = files->get(list->item(i)->text(),"WAD");
		if (wadfile.isNull()) return (files->lastError());

		QFileInfo inf = QFileInfo(wadfile);
		if (!inf.isReadable()) return (i18n("The PWAD file '<b>%1</b>' was not found or is not readable").arg(wadfile));
		if (first) cmd += "-file";
		cmd += wadfile;
		first = false;
	}

	return (QString::null);
}


QString GenericGame::playLoadGame(QStringList &cmd,const QString &load)
{
	cmd += "-loadgame";
	cmd += load;

	return (QString::null);
}


QString GenericGame::playPlayDemo(QStringList &cmd,RemoteFiles *files,const KURL &lmp)
{
	const QString df = demoFormat();		// demo format for this game
	if (df.isNull()) return ("!");			// not supported, should never happen

	QString lmpfile = files->get(lmp,df);
	if (lmpfile.isNull()) return (files->lastError());

	QFileInfo inf = QFileInfo(lmpfile);
	if (!inf.isReadable()) return (i18n("The %1 file '<b>%2</b>' was not found or is not readable").arg(df,lmpfile));

	cmd += "-playdemo"; cmd += lmpfile;
	return (QString::null);
}


QString GenericGame::playStartLevel(QStringList &cmd,const QString &warp,int skill)
{
	cmd += "-warp";
	if (warp.startsWith("E"))			// Doom 1 level format
	{
		int e,m;
		sscanf(warp.upper().ascii(),"E%dM%d:",&e,&m);
		cmd += QString::number(e);
		cmd += QString::number(m);
	}
	else cmd += warp;				// plain Doom 2 format

	cmd += "-skill"; cmd += QString::number(skill);
	return (QString::null);
}


QString GenericGame::optionsEnemies(QStringList &cmd)
{
	if (enemiesoption & EnemiesOptions::None) cmd += "-nomonsters";
	else
	{
		if (enemiesoption & EnemiesOptions::Respawn) cmd += "-respawn";
		if (enemiesoption & EnemiesOptions::Fast) cmd += "-fast";
	}

	return (QString::null);
}


QString GenericGame::optionsSound(QStringList &cmd)
{
	if (soundoption==SoundOptions::NoSound) cmd += "-nosound";
	else
	{
		if (soundoption & SoundOptions::NoSfx) cmd += "-nosfx";
		if (soundoption & SoundOptions::NoMusic) cmd += "-nomusic";
	}
	return (QString::null);
}


QString GenericGame::getPlayCommand(QStringList &cmd)
{
	kdDebug(0) << k_funcinfo << endl;

	QString err = QString::null;
	RemoteFiles *files = RemoteFiles::self();

	if (!(err = getMainWad(cmd,files,mainwad->url())).isNull()) return (err);
	if (!(err = getPWads(cmd,files,pwads->listBox())).isNull()) return (err);

	QString load = loadgame->currentText();		// first priority: load saved
	int pos = load.find(':');			// format is "index: details"
	if (pos>0)	
	{
		if (!(err = playLoadGame(cmd,load.left(pos))).isNull()) return (err);
	}
	else
	{
		QString lmp = demolmp->url();
		if (!lmp.isEmpty())			// next priority: play demo
		{
			if (!(err = playPlayDemo(cmd,files,lmp)).isNull()) return (err);
		}
		else
		{
			QListBoxItem *selwarp = startlevel->selectedItem();
			if (selwarp!=NULL)		// lowest priority: start level
			{
				QString seltext = selwarp->text();
				int p = seltext.find(':');
				if (p>=0)
				{
					if (!(err = playStartLevel(cmd,seltext.left(p),skill->currentItem())).isNull()) return (err);
				}
			}
		}
	}

	if (!(err = optionsDisplay(cmd)).isNull()) return (err);
	if (!(err = optionsEnemies(cmd)).isNull()) return (err);
	if (!(err = optionsSound(cmd)).isNull()) return (err);

	cmd += ListParser::parseToList(extraargs);	// add on custom options
	return (QString::null);				// everything OK
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Saved game and demo support						//
//									//
//////////////////////////////////////////////////////////////////////////

void GenericGame::checkForNewSaves()
{
	int c = loadgame->currentItem();
	loadgame->clear();
	loadgame->insertItem(i18n("(None)"));
	loadgame->insertStringList(findSavedGames());
	if (loadgame->count()<2) loadgame->setEnabled(false);
	if (c<loadgame->count()) loadgame->setCurrentItem(c);

	if (loadsaved>=0)				// read from config file
	{
		loadgame->setCurrentItem(loadsaved);
		slotOptionsChanged();
		loadsaved = -1;				// don't use it again
	}
}


QString GenericGame::demoFormat() const			// the original LMP format
{
	return ("LMP");
}
