/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  InfoDisplay								//
//  -----------								//
//									//
//  Display an information file associated with a WAD or demo.		//
//  This is normally a plain text file, although given the use of	//
//  a generic KPart for display any type of file could potentially	//
//  be shown.								//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <klocale.h>
#include <kdialog.h>
#include <kurl.h>
#include <kapplication.h>
#include <kmessagebox.h>
#include <kmimetype.h>
#include <kprotocolinfo.h>

#include <kio/netaccess.h>
#include <kio/jobclasses.h>
#include <kio/job.h>

#include <kparts/componentfactory.h>

#include <qlayout.h>
#include <qvbox.h>

#include "remotefiles.h"
#include "matchresolver.h"

#include "infodisplay.h"
#include "infodisplay.moc"

//////////////////////////////////////////////////////////////////////////
//									//
//  File access								//
//									//
//  Using a pessimistic remote test here (2nd argument to		//
//  KIO::NetAccess::exists() set to 'false').  See documentation of	//
//  KIO::Stat for more information.					//
//									//
//////////////////////////////////////////////////////////////////////////

bool InfoDisplay::tryExists(const KURL &u)
{
	kdDebug() << k_funcinfo << "trying " << u << endl;

	QString local = RemoteFiles::isAccessibleFile(u);
	if (!local.isNull())				// see if a local file
	{						// yes, can look at it directly
		QFile f(local);
		if (!f.exists()) return (false);
		kdDebug() << "  found local " << f.name() << endl;
		url = KURL::fromPathOrURL(local);	// note this for showing
	}
	else						// not a local file
	{						// ask protocol if it exists
		if (!KIO::NetAccess::exists(u,false,kapp->mainWidget())) return (false);
		kdDebug() << "  found remote " << u << endl;
		url = u;				// note this for showing
	}

	return (true);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Directory listing							//
//									//
//////////////////////////////////////////////////////////////////////////

void InfoDisplay::slotListingEntries(KIO::Job *job,const KIO::UDSEntryList &list)
{
	kdDebug(0) << k_funcinfo << "received list of " << list.count() << endl;

	for (KIO::UDSEntryList::ConstIterator it = list.begin(); it!=list.end(); ++it)
	{						// loop over received entries
		for (KIO::UDSEntry::const_iterator et = (*it).begin(); et!=(*it).end(); ++et) 
		{					// loop over entry's atoms
			if ((*et).m_uds==KIO::UDS_NAME)	// found atom for file name,
			{				// save it in list
				listingEntries += (*et).m_str;
				break;			// the only interesting atom
			}
		}
	}
}


void InfoDisplay::slotListingFinished(KIO::Job *job)
{
	kdDebug() << k_funcinfo << " with list of " << listingEntries.count() << endl;

	if (job->error())				// error during the listing
	{
		job->showErrorDialog(this);
		slotFinished();
		return;
	}

	url.setDirectory(url.path());			// ensure trailing '/' present

	MatchResolver res(src,"TXT");			// resolve match in list
	MatchResolver::Status matchState = res.match(listingEntries,
						     i18n("<qt>Select information file for <b>%1</b>:").arg(src.fileName()));
	if (matchState==MatchResolver::Cancelled) return;
							// abort if cancelled
	if (matchState<MatchResolver::Extension)	// no suitable match found
	{
		url = KURL();				// clear for error message later
	}
	else						// something found or user selected
	{
		QString match = res.found();
		if (matchState==MatchResolver::Remember) RemoteFiles::self()->saveMatch(src,"TXT",match);
		url.setFileName(match);			// use the user selection
	}

	kdDebug(0) << "  resolved to " << url.prettyURL() << endl;
	needsListing = false;				// don't call us again!
	showInfo();					// show the selected result
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

InfoDisplay::InfoDisplay(const KURL &s,QWidget *parent)
	: KDialogBase(parent,NULL,false,QString::null,KDialogBase::Close)
{
	kdDebug(0) << k_funcinfo << "for " << s << endl;

	src = s;					// save for later
	m_part = NULL;					// not created yet
	m_widget = NULL;
	needsListing = false;				// not yet, anyway

	KURL u = src;					// copy, we'll try variations
							// original URL without extension
	QString base = RemoteFiles::nameBase(src.fileName());

	QString match = RemoteFiles::self()->cachedMatch(src,"TXT");
	if (!match.isNull())				// see if match is cached
	{
		u.setFileName(match);
		if (tryExists(u)) return;		// use that if found
	}

	u.setFileName(base+".txt");			// look for corresponding ".txt"
	if (tryExists(u)) return;			// use that if found

	u.setFileName(base+".TXT");			// look for corresponding ".TXT"
	if (tryExists(u)) return;			// use that if found

	KURL arc = RemoteFiles::self()->isArchiveFile(src);
	if (arc.isValid())				// see if an archive file
	{						// yes, try looking inside it
		QString localFile = RemoteFiles::self()->get(arc,"TXT");
		if (!localFile.isNull())		// found inside archive
		{
			url = KURL::fromPathOrURL(localFile);
			return;				// use the extracted file
		}
	}

	if (KProtocolInfo::supportsListing(src))	// can we list associated files?
	{
		u.setFileName(QString::null);		// reduce to directory only
		url = u;				// note for listing later
		needsListing = true;			// note this needs doing
	}

	kdDebug(0) << k_funcinfo << "done url=" << url.prettyURL() << " needsListing=" << needsListing << endl;
}


InfoDisplay::~InfoDisplay()
{
	saveDialogSize("InfoDisplay");
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Action slots							//
//									//
//////////////////////////////////////////////////////////////////////////

void InfoDisplay::slotFinished()
{
	if (m_part!=NULL) delete m_part;
	m_part = NULL;
	delayedDestruct();
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Display the information						//
//									//
//////////////////////////////////////////////////////////////////////////

bool InfoDisplay::showInfo()
{
	if (!url.isValid())
	{
		KMessageBox::sorry(NULL,QString("<qt>No information is available for<br><b>%1</b>").arg(src.prettyURL()));
		slotFinished();				// we'll just go away then
		return (false);
	}

	if (needsListing)				// a listing is needed...
	{
		listingEntries.clear();			// clear matched list
		KIO::ListJob *job = KIO::listDir(url,true,false);
		connect(job,SIGNAL(entries(KIO::Job *,const KIO::UDSEntryList &)),
			this,SLOT(slotListingEntries(KIO::Job *,const KIO::UDSEntryList &)));
		connect(job,SIGNAL(result(KIO::Job *)),
			this,SLOT(slotListingFinished(KIO::Job *)));
		return (true);				// call us again when done
	}

	m_widget = new QVBox(this);
	m_widget->layout()->setSpacing(KDialog::marginHint());
	connect(this,SIGNAL(finished()),this,SLOT(slotFinished()));
	setMainWidget(m_widget);

	QSize size = configDialogSize("InfoDisplay");
	if (size.width()<200) size = QSize(560,400);
	setInitialSize(size);

	KMimeType::Ptr mimetype = KMimeType::findByURL(url);
	m_part = KParts::ComponentFactory::createPartInstanceFromQuery<KParts::ReadOnlyPart>(mimetype->name(),QString::null,m_widget,0,this);
	if (m_part==NULL)				// can't find part for content
	{
		slotFinished();				// we'll just go away then
		return (false);
	}

	m_part->openURL(url);				// load file into part

	setCaption(kapp->makeStdCaption(i18n("Information for %1 (%2)").arg(src.fileName(),url.fileName())));
	show();
	return (true);
}
