/////////////////////////////////////////////////// -*- mode:c++; -*- ////
//  Copyright (C) 2008 by Jonathan Marten				//
//  jjm@keelhaul.me.uk							//
//									//
//  This program is free software; you can redistribute it and/or	//
//  modify it under the terms of the GNU General Public License as	//
//  published by the Free Software Foundation; either version 2 of	//
//  the License, or (at your option) any later version.			//
//									//
//  This program is distributed in the hope that it will be useful,	//
//  but WITHOUT ANY WARRANTY; without even the implied warranty of	//
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	//
//  GNU General Public License for more details.			//
//									//
//  You should have received a copy of the GNU General Public License	//
//  along with this program; if not, write to the			//
//  Free Software Foundation, Inc.,					//
//  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.		//
//									//
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//									//
//  This file is part of Kadoom, a KDE desktop front end to various	//
//  Doom game engines.  See http://www.keelhaul.me.uk/linux/kadoom	//
//  for the latest version and further information.			//
//									//
//									//
//  SetupWizard								//
//  -----------								//
//									//
//  Select and configure a game engine, step by step.			//
//									//
//////////////////////////////////////////////////////////////////////////

#include <kdebug.h>
#include <kglobal.h>
#include <klocale.h>
#include <kwizard.h>
#include <kdialog.h>
#include <kurlrequester.h>
#include <kfiledialog.h>
#include <kmessagebox.h>
#include <kactivelabel.h>

#include <qlabel.h>
#include <qlayout.h>
#include <qlistbox.h>
#include <qwidget.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qpushbutton.h>

#include "gamemanager.h"

#include "setupwizard.h"
#include "setupwizard.moc"

//////////////////////////////////////////////////////////////////////////
//									//
//  Text strings							//
//									//
//////////////////////////////////////////////////////////////////////////

static const QString page1Acaption = i18n("Available Engines");
static const QString page1Bcaption = i18n("Select Engine");
static const QString page2caption = i18n("Engine Executable");
static const QString page3caption = i18n("WAD Files");
static const QString page4caption = i18n("Finished");

static const QString page1Atext = i18n("<qt><p>A Doom engine is needed to play the game. \
There are many different engines available, with various graphics and gameplay options.\
<p>\
These engines were found on your system, select one and use <b>Next</b> to go to the next page. \
If none are listed or if an engine is installed but in a non-standard location, select \"Other\".");

static const QString page1Btext = i18n("<qt><p>A Doom engine needs to be installed to play the game. \
These are all the engines that are supported by this application.\
<p>\
If one of these engines is installed but in a non-standard location, select it and use <b>Next</b> \
to locate it on the next page.\
<p>\
If you do not have any engine installed, select any one from the list for information about it. \
Install a suitable engine on your system before continuing.");
static const QString page2text = i18n("<qt><p>Specify the location of the <b>%1</b> engine executable. \
<p>\
You only need to select or change this if the engine is installed in a non-standard location.");
static const QString page3text = i18n("<qt><p>A WAD file is needed to play the game; \
this contains all of the maps, graphics and sounds.  You can use the file from a purchased copy of \
one of the Doom series games, or if you do not have any of these you can download the shareware \
version (to play the first episode of Doom&nbsp;1 only) from \
<a href=\"ftp://ftp.idsoftware.com/idstuff/doom/win95/doom95.zip\">ftp://ftp.idsoftware.com/idstuff/doom/win95/doom95.zip</a> \
and extract the <b>DOOM1.WAD</b> file from that.  You do not need Windows&trade;&nbsp;95 to use this!\
<p>\
You could also use the full WAD from the FreeDoom project, see \
<a href=\"http://freedoom.sourceforge.net/\">http://freedoom.sourceforge.net</a> \
for more information.");
static const QString page4text = i18n("<qt>The engine and file locations are now configured. \
Use <b>Finish</b> to save the configuration and set up to start playing.");

//////////////////////////////////////////////////////////////////////////
//									//
//  Constructor/destructor						//
//									//
//////////////////////////////////////////////////////////////////////////

SetupWizard::SetupWizard(const QString &title,QWidget *parent)
	: KWizard(parent)
{
	kdDebug(0) << k_funcinfo << endl;

	game = NULL;					// none selected yet
	manager = GameManager::self();			// we'll use this a lot

	setCaption(title);
	setupPage1A();
	setupPage1B();
	setupPage2();
	setupPage3();
	setupPage4();
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Moving between pages						//
//									//
//////////////////////////////////////////////////////////////////////////

void SetupWizard::showPage(QWidget *page)
{
	KWizard::showPage(page);

	if (page==page1A || page==page1B) slotPage1EngineSelected();
	else if (page==page2)
	{
		if (game!=NULL)
		{
			page2info->setText(page2text.arg(game->name));
			page2exec->setURL(manager->findExecFor(game));
		}
		slotPage2ExecSelected();
	}
	else if (page==page3)
	{
		if (game!=NULL) page3wad->setURL(manager->findDataFor(game));
		slotPage3WadSelected();
	}
}


void SetupWizard::next()
{
	bool ok = true;

	if (currentPage()==page2)
	{
		const QString u = page2exec->url();
		ok = (!u.isEmpty());
		if (ok)
		{
			QFileInfo fi(u);
			if (!fi.exists())
			{
				KMessageBox::sorry(this,i18n("<qt>The file <b>%1</b><br>does not exist.").arg(fi.absFilePath()));
				ok = false;
			}
			else if (!fi.isExecutable())
			{
				KMessageBox::sorry(this,i18n("<qt>The file <b>%1</b><br>exists but is not executable.").arg(fi.absFilePath()));
				ok = false;
			}

			if (ok && fi.isRelative())
			{
				fi.convertToAbs();
				page2exec->setURL(fi.filePath());
			}
		}
	}
	else if (currentPage()==page3)
	{
		const QString u = page3wad->url();
		ok = (!u.isEmpty());
		if (ok)
		{
			QFileInfo fi(u);
			if (!fi.exists())
			{
				KMessageBox::sorry(this,i18n("<qt>The directory <b>%1</b><br>does not exist.").arg(fi.absFilePath()));
				ok = false;
			}
			else if (!fi.isDir())
			{
				KMessageBox::sorry(this,i18n("<qt>The location <b>%1</b><br>is not a directory.").arg(fi.absFilePath()));
				ok = false;
			}

			if (ok && fi.isRelative())
			{
				fi.convertToAbs();
				page3wad->setURL(fi.filePath());
			}
		}
	}

	if (ok) KWizard::next();
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Page 1:  list of engines						//
//									//
//////////////////////////////////////////////////////////////////////////

void SetupWizard::setupPage1A()
{
	page1A = new QWidget(this);

	QHBoxLayout *l = new QHBoxLayout(page1A,KDialog::marginHint(),KDialog::spacingHint());

	page1Alist = new QListBox(page1A);
	page1Alist->setSelectionMode(QListBox::Single);
	page1Alist->setHScrollBarMode(QScrollView::AlwaysOff);
	page1Alist->setVScrollBarMode(QScrollView::AlwaysOn);
	page1Alist->setMinimumHeight(80);
	connect(page1Alist,SIGNAL(selectionChanged()),this,SLOT(slotPage1EngineSelected()));
	l->addWidget(page1Alist);

	for (const GameManager::gameInfo *it = manager->firstInfo(); it!=NULL; it = manager->nextInfo())
	{
		QString s = it->name;
		QString e = manager->findExecFor(it);
		if (!e.isNull()) page1Alist->insertItem(s);
	}
	page1Alist->insertItem(i18n("(Other)"));

	page1Alist->setMinimumWidth(page1Alist->maxItemWidth()+page1Alist->verticalScrollBar()->width()+20);

	page1Ainfo = new KActiveLabel(page1Atext+"<p>"+i18n("Select an engine for more information about it."),page1A);
	l->addWidget(page1Ainfo,1);

	page1A->setMinimumHeight(240);
	addPage(page1A,page1Acaption);
	setHelpEnabled(page1A,false);
}


void SetupWizard::setupPage1B()
{
	page1B = new QWidget(this);
	QHBoxLayout *l = new QHBoxLayout(page1B,KDialog::marginHint(),KDialog::spacingHint());

	page1Blist = new QListBox(page1B);
	page1Blist->setSelectionMode(QListBox::Single);
	page1Blist->setHScrollBarMode(QScrollView::AlwaysOff);
	page1Blist->setVScrollBarMode(QScrollView::AlwaysOn);
	page1Blist->setMinimumHeight(80);
	connect(page1Blist,SIGNAL(selectionChanged()),this,SLOT(slotPage1EngineSelected()));
	l->addWidget(page1Blist,1);

	for (const GameManager::gameInfo *it = manager->firstInfo(); it!=NULL; it = manager->nextInfo())
	{
		page1Blist->insertItem(it->name);
	}
	page1Blist->setMinimumWidth(page1Blist->maxItemWidth()+page1Blist->verticalScrollBar()->width()+20);

	page1Binfo = new KActiveLabel(page1Btext+"<p>"+i18n("Select an engine for more information about it."),page1B);
	l->addWidget(page1Binfo,1);

	addPage(page1B,page1Bcaption);
	setHelpEnabled(page1B,false);
	nextButton()->setDefault(true);
}


void SetupWizard::slotPage1EngineSelected()
{
	if (currentPage()==page1A)
	{
		QString t = page1Atext;

		QListBoxItem *item = page1Alist->selectedItem();
		QString g = (item==NULL ? QString::null : item->text());
		if (!g.startsWith("("))			// not selected "Other"
		{
			game = manager->findNamed(g);
			if (game!=NULL)
			{
				setAppropriate(page1B,false);

				if (!game->url.isNull())
				{
					t += "<p>"+i18n("See <a href=\"%1\">%2</a> for more information on <b>%3</b>.").arg(game->url,game->url,game->name);
				}

				if (!game->notes.isNull())
				{
					t += "<p>"+game->notes;
				}
				else
				{
					if (game->url.isNull())
					{
						t += "<p>"+i18n("No further information is available for <b>%1</b>.").arg(game->name);
					}
				}
			}
		}
		else
		{
			game = NULL;
			setAppropriate(page1B,true);
			t += "<p>"+i18n("Choose an engine on the next page.");

		}

		page1Ainfo->setText(t);
		setNextEnabled(page1A,!g.isNull());
	}
	else
	{
		QString t = page1Btext;

		QListBoxItem *item = page1Blist->selectedItem();
		QString g = (item==NULL ? QString::null : item->text());
		game = manager->findNamed(g);
		if (game!=NULL)
		{
			if (!game->url.isNull())
			{
				t += "<p>"+i18n("See <a href=\"%1\">%2</a> for more information on <b>%3</b>.").arg(game->url,game->url,game->name);
			}

			if (!game->notes.isNull())
			{
				t += "<p>"+game->notes;
			}
			else
			{
				if (game->url.isNull())
				{
					t += "<p>"+i18n("No further information is available for <b>%1</b>.").arg(game->name);
				}
			}
		}

		page1Binfo->setText(t);
		setNextEnabled(page1B,game!=NULL);
	}
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Page 2:  engine executable						//
//									//
//////////////////////////////////////////////////////////////////////////

void SetupWizard::setupPage2()
{
	page2 = new QWidget(this);
	QVBoxLayout *l = new QVBoxLayout(page2,KDialog::marginHint(),KDialog::spacingHint());

	page2info = new QLabel(page2text.arg(""),page2);
	l->addWidget(page2info,1,Qt::AlignTop);

	page2exec = new KURLRequester(page2);
	page2exec->setMode(KFile::File+KFile::ExistingOnly+KFile::LocalOnly);

	// TODO: can we filter for executables?
	page2exec->fileDialog()->setCaption(page2caption);
	connect(page2exec,SIGNAL(textChanged(const QString &)),this,SLOT(slotPage2ExecSelected()));

	QLabel *lab = new QLabel(i18n("Executable location:"),page2);
	lab->setBuddy(page2exec);
	l->addWidget(lab,1,Qt::AlignBottom);
	l->addWidget(page2exec,0,Qt::AlignTop);

	addPage(page2,page2caption);
	setHelpEnabled(page2,false);
	nextButton()->setDefault(true);
}


void SetupWizard::slotPage2ExecSelected()
{
	kdDebug(0) << k_funcinfo << " url=" << page2exec->url() << endl;
	setNextEnabled(page2,!page2exec->url().isEmpty());
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Page 3:  WAD files							//
//									//
//////////////////////////////////////////////////////////////////////////

void SetupWizard::setupPage3()
{
	page3 = new QWidget(this);
	QVBoxLayout *l = new QVBoxLayout(page3,KDialog::marginHint(),KDialog::spacingHint());

	KActiveLabel *info = new KActiveLabel(page3text,page3);
	l->addWidget(info,1,Qt::AlignTop);

	page3wad = new KURLRequester(page3);
	page3wad->setMode(KFile::Directory+KFile::ExistingOnly+KFile::LocalOnly);
	page3wad->fileDialog()->setCaption(page3caption);
	connect(page3wad,SIGNAL(textChanged(const QString &)),this,SLOT(slotPage3WadSelected()));

	QLabel *lab = new QLabel(i18n("WAD files location (directory):"),page3);
	lab->setBuddy(page3wad);
	l->addWidget(lab,1,Qt::AlignBottom);
	l->addWidget(page3wad,0,Qt::AlignTop);

	addPage(page3,page3caption);
	setHelpEnabled(page3,false);
	nextButton()->setDefault(true);
}


void SetupWizard::slotPage3WadSelected()
{
	kdDebug(0) << k_funcinfo << " url=" << page3wad->url() << endl;
	setNextEnabled(page3,!page3wad->url().isEmpty());
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Page 4:  finished							//
//									//
//////////////////////////////////////////////////////////////////////////

void SetupWizard::setupPage4()
{
	page4 = new QWidget(this);
	QVBoxLayout *l = new QVBoxLayout(page4,KDialog::marginHint(),KDialog::spacingHint());

	QLabel *info = new QLabel(page4text,page4);
	l->addWidget(info,1,Qt::AlignTop);

	addPage(page4,page4caption);
	setHelpEnabled(page4,false);
	setFinishEnabled(page4,true);
	finishButton()->setDefault(true);
}

//////////////////////////////////////////////////////////////////////////
//									//
//  Returned results							//
//									//
//////////////////////////////////////////////////////////////////////////

const QString SetupWizard::getGameKey() const
{
	return (game!=NULL ? game->key : NULL); 
}


const QString SetupWizard::getGameName() const
{
	return (game!=NULL ? game->name : NULL); 
}


const QString SetupWizard::getExecutable() const
{
	return (page2exec->url());
}


const QString SetupWizard::getDataDir() const
{
	return (page3wad->url());
}
